

################################################################################


class Geometry:
  '''
  initialize with one of the following ways:
  A) geometry [string], e.g., "mesh.vol" (some netgen mesh)
     meshSize=None
     scaling [float], e.g., 1e-9 (nm)
     structuedMesh=False
     --> the previosly saved mesh "mesh.vol" will be loaded and used
  B) geometry [standardgeometry], e.g., Ball, Cuboid, Disk, UnitBall, UnitCube
     meshSize [float]
     scaling [float], e.g., 1e-9 (nm)
     structuedMesh=False
     --> The parameters will be passed to Netgen where an unstructured 
      mesh will be generated
  C) geometry=Cuboid(...)
     meshSize [(float1, float2, float3)]
     scaling [float], e.g., 1e-9 (nm)
     structuedMesh=True
     ---> you will get a manually generated mesh:
      The Cuboid is divided into smaller cuboids of 
      dimensions float1 x float2 x float3.
      Each of the small cuboids will be split into 6 tetrahedra
      The mesh will have no dihedral angle greater than pi/2
      The mesh-size will be sqrt(float1^2 + float2^2 + float3^2)
  D) geometry [netgen.csg.CSGeometry] ... some CSGeometry you defined using netgen.csg
     IMPORTANT: use the name "bc_dirichlet" for all boundaries!
     (otherwise the stray-field computations will be incorrect)
     meshSize [float]
     scaling [float], e.g., 1e-9 (nm)
     structuedMesh=False
  in B+D) force the maximum meshSize requirement by
      forceNetgenMeshSize=True

  NOTE:
    The scaling parameter seems artificial. Indeed it is.
    But Netgen, for some reason, is not capable of meshing geometries 
    on very small or very large scales - I have no idea why...
    So, if for instance you want a cube with edge length=1e-9:
    use Cuboid((0,0,0), (1,1,1)) + scaling=1e-9 rather 
    than Cuboid((0,0,0), (1e-9, 1e-9, 1e-9)) + scaling=1
  '''
  
  def __init__(self, geometry=None, meshSize=None, forceNetgenMeshSize=False, scaling=1.0 \
    , structuredMesh=False, structuredObtuseMesh=False):
    
    self._geometry = geometry
    self.meshSize = meshSize
    self.forceNetgenMeshSize = forceNetgenMeshSize
    self.scaling = scaling

    self.structuredMesh = structuredMesh
    self.structuredObtuseMesh = structuredObtuseMesh
    
    self._mesh = None
    

#------------------------------------------------------------------------------#

  
  def GetReady(self):
    import sys
    from netgen.csg import CSGeometry
    from ngsolve import Mesh as ngsolveMesh
    from netgen.meshing import Mesh as netgenMesh
    from commics._setup.geometries.standardGeometries import _standardGeometry
    from ngsolve import ngsglobals

    ngsglobals.msg_level = 0

    if not(self._mesh == None):
      print("From Geometry.GetReady(): Nothing to do, mesh is ready.")
      return

    if(isinstance(self._geometry, str)):
      if(self.meshSize != None):
        print("Error in Geometry.GetReady():")
        print("meshSize must not be set, if mesh is loaded from file!!!")
        print(" ---> Terminate.")
        sys.exit()

      ngmesh = netgenMesh(3)
      ngmesh.Load( self._geometry )
    else:
      if(self.structuredMesh or self.structuredObtuseMesh):
        ngmesh = self._GenerateStructuredMesh()
      else:
        if( isinstance(self._geometry, _standardGeometry._StandardGeometry)):
          ngGeo = self._geometry.GetNgSolveGeometry()
        else:
          ngGeo = self._geometry
  
        if(self.meshSize == None):
          print("Error in Geometry.GetReady():")
          print("Specify meshSize --> Terminate.")
          sys.exit()
      
        geo = CSGeometry()
        geo.Add(ngGeo)
        
        if self.forceNetgenMeshSize:
          from commics._tools.misc import MaximumMeshSize
          achievedMeshSize = False
          j = 1
          print("Netgen meshing ........ Desired meshSize =", self.meshSize) 
          while not(achievedMeshSize):
            ngmesh = geo.GenerateMesh( maxh = self.meshSize / j )
            actualMeshSize = MaximumMeshSize(ngmesh)
            achievedMeshSize = actualMeshSize <= self.meshSize
            print("Itaration Nr.%i: Desired meshSize = %g --- Actual meshSize = %g" % (j, self.meshSize/j, actualMeshSize))
            j += 1
          print("Appropriate mesh was generated ........") 
        else:
          ngmesh = geo.GenerateMesh( maxh = self.meshSize )
      
    self._mesh = ngsolveMesh( ngmesh )
  
  
#------------------------------------------------------------------------------#


  def _GenerateStructuredMesh(self):
    from commics._setup.geometries.standardGeometries import cuboid
    from commics._setup.geometries.standardGeometries._structuredMeshing \
      import StructuredMesh_Cuboid, StructuredObtuseMesh_Cuboid
    if(len(self.meshSize) != 3):
      raise ValueError("For structured mesh, provide meshSize as (hx, hy, hz)")
    if not(isinstance(self._geometry, cuboid.Cuboid)):
      raise NotImplementedError("Structured meshing only for Cuboid geo")

    if self.structuredObtuseMesh and self.structuredMesh:
      raise ValueError("structuredObtuseMesh=True and structuredMesh=True ?!")

    p1, p2 = self._geometry.GetXminXmax()

    if self.structuredMesh:
      return StructuredMesh_Cuboid(p1, p2, self.meshSize)
    else:
      return StructuredObtuseMesh_Cuboid(p1, p2, self.meshSize)


#------------------------------------------------------------------------------#

  
  def __str__(self):
    string = "\n\n ----- " + self.__class__.__name__ + " ----- \n\n"
    if not(self._mesh == None):
      string += "Vertices: " + str(self._mesh.nv) + "\n"
      string += "Elements: " + str(self._mesh.ne) + "\n"
      string += "meshSize = " + str(self.meshSize) + "\n"
      string += "forceNetgenMeshSize = " + str(self.forceNetgenMeshSize) + "\n"
      string += "scaling = " + str(self.scaling) +"\n"
      string += "structuredMesh = " + str(self.structuredMesh) +"\n"

      if isinstance(self._geometry, str):
        string += "file = " + str(self._geometry) + "\n"

      return string
    else:
      string = "From Geometry.__str__(): Mesh not ready: Call GetReady() first!"
  
    return string

  
#------------------------------------------------------------------------------#

  
  def GetMesh(self):
    if not(self._mesh == None):
      return self._mesh
    else:
      print("From Geometry.GetMesh() : \
        Mesh not ready: Call GetReady() first!")
  
  
#------------------------------------------------------------------------------#

  
  def Save(self, filename):
    if not(self._mesh == None):
      ext = ".vol"
      if(filename[-4:] == ext):
        ext = ""
      self._mesh.ngmesh.Save( filename + ext )
    else:
      print("From Geometry.Save(): \
        Mesh not ready: Call GetReady() first!")
  
  
#------------------------------------------------------------------------------#


################################################################################
