### class Parameters ###


################################################################################


class Parameters:
  '''
  Objects of this class are meant to specify
    - material parameters [in SI units]
      .A ... exchange stiffness
      .Ms ... saturation magnetization
      .K ... anisotropy constant
      .D ... DMI constant
      .alpha ... Gilber damping parameters
      .P ... polarization rate (if current applied)
      .xi ... zhang li non-adiabaticity
      .Je ... applied current (1-dim for Slonczewski, 3-dim for Zhang-Li),
        each component may be (t, x, y, z)-dependent (e.g., lambda fct)
      .d ... thickness of geometry (necessary if DMI or Slonczewski used)
      .p ... current polarization (Slonczewski)
    - simulation options:
      .m0 ... Initial magnetization: ["file.vtk", (0, 0, 1), 
        or (ngsolveCF, ngolvesCF, ngsolveCF)] are possible
      .theta ... degree of implicitness, available for, e.g., TPS1
      .T_start, .T_end ..., e.g., 0 and 3e-9 to simulate for 3ns
    - physical constants [those are predefined, i.e., they have default values]
      .gamma0 [2.212761569e+05]
      .mu0 [1.2566370614359173e-06]
      .g [2.0023193043617]
      .muB [9.27400968e-24]
      .e [1.6021766208e-19]
      .hbar [1.054571800e-34]
    - which lower-order contributions to consider:
      .H_ext = [ (f,g,h) where f, g, h are either constants 
                 or functions in (t, x, y, z); with t in physical seconds; 
                 (x, y, z) w.r.t. to the NgSolve geometry, i.e., unphysical dimension]
      .easy_axis [(1,0,0)] and .K for uniaxial anisotropy
      .maxwellCoupling = ['none', 'magnetostaticField', 'strayField','oerstedField']
      (magnetostaticField = strayfield & oersted field)
      .dmCoupling = ['none', 'bulk', 'interfacial'] (if interfacial, .dmi_axis (default = 'z'))
      .spintronicsCoupling = ['none', 'zhangLi', 'slonczewski']
      .externalFieldRecordingPoint = (x1, x2, x3) in Omega, default "first mesh point"
  '''
  

  def __init__(self, A=None, Ms=None, K=0.0, D=0.0, dmi_axis='z', gamma0=2.212761569e+05, mu0=1.2566370614359173e-06, alpha=1.0, T_start=None, T_end=None, timeStepSize=None, H_ext=None, easy_axis=(1,0,0), g=2.0023193043617, muB=9.27400968e-24, e=1.6021766208e-19, P=None, xi=None, dmCoupling="none", dmImplicitTreatment=False, maxwellCoupling="strayField", spintronicsCoupling="none", slonczewskiOutOfPlane_axis='z', hbar=1.054571800e-34, Je=None, flag_Je_stationary=False, d=None, p=(1,0,0), TOL=1.0e-8, preconditionerType="jacobi", alphaPreconditioner=None, QmatType="householder", QmatAxis="e3", guessType="old", theta=0.5, m0=None, externalFieldRecordingPoint="first mesh point"):
    self.A = A
    self.Ms = Ms
    self.K = K
    self.D = D
    self.dmi_axis = dmi_axis

    self.gamma0 = gamma0
    self.mu0 = mu0
    
    self.alpha = alpha

    self.T_start = T_start
    self.T_end = T_end
    self.timeStepSize = timeStepSize
    
    self.H_ext = H_ext
    self.easy_axis = easy_axis
    
    self.dmCoupling = dmCoupling
    self.dmImplicitTreatment = dmImplicitTreatment
    self.maxwellCoupling = maxwellCoupling
    
    # spintronics par
    self.spintronicsCoupling = spintronicsCoupling
    self.slonczewskiOutOfPlane_axis = slonczewskiOutOfPlane_axis
    self.Je = Je
    self.flag_Je_stationary = flag_Je_stationary
    self.e = e
    # zhangLi par
    self.P = P
    self.g = g
    self.muB = muB
    self.xi = xi
    # slonczewski par
    self.hbar = hbar
    self.d = d
    self.p = p

    self.TOL = TOL
    self.preconditionerType = preconditionerType
    self.alphaPreconditioner = alphaPreconditioner
    self.guessType = guessType
    self.QmatType = QmatType
    self.QmatAxis = QmatAxis

    self.theta = theta

    self.m0 = m0

    self.externalFieldRecordingPoint = externalFieldRecordingPoint

#------------------------------------------------------------------------------#


  def __str__(self):
    string = "\n\n ----- " + self.__class__.__name__ + " ----- \n\n"
    for key, value in self.__dict__.items():
      if not(key.startswith("_")):
        string += key + " = " + str(value) + "\n"
    
    return string


#------------------------------------------------------------------------------#


  def DeepCopy(self): 

    copy = type(self)()
    for key, value in self.__dict__.items():
      copy.__setattr__(key, value)

    return copy


#------------------------------------------------------------------------------#


################################################################################
