## a Wrapper around the commics module
# provides easy access to experimental convergence orders


################################################################################


class ConvergenceOrderWrapper:
  
  def __init__(self, referenceScheme, otherScheme, nrSimulations, geometry, parameters=None, otherParameters=None, output=None, numthreads=1):
    from commics import Integrator, Parameters

    self.R = nrSimulations
    
    self.foldername = "convergenceOrderResult"
    self.filenameParameters = "parameters.txt"
    self.header = "Convergence order simulation"
    
    self._firstRun = True
    self._integrators = None
    self._referenceSimulation = Integrator(referenceScheme, geometry, parameters, output, numthreads)
    self._otherScheme = otherScheme
    self._otherParameters = parameters.DeepCopy() if otherParameters == None else otherParameters
    if self._otherParameters == None:
      self._otherParameters = Parameters()

    self._refTotalSteps = 0
    
    self._evolutionName = "Evolution"
    self._errL2filename = "errL2"
    self._errH1semifilename = "errH1semi"
    self._errH1filename = "errH1"
    self._errL2FileHeader = ["t", "max", "now"]
    self._errH1semiFileHeader = ["columns: t", "max", "now"]
    self._errH1FileHeader = ["columns: t", "max", "now"]
    self._columnWidth = self._referenceSimulation._integrator.output.columnWidth
    self._shellColumnWidth = 14
    self._ratesL2filename = "ratesL2"
    self._ratesH1semifilename = "ratesH1semi"
    self._ratesH1filename = "ratesH1"
    
    self._errL2FileHeadline = "L2-Errors"
    self._errH1semiFileHeadline = "H1semi-Errors"
    self._errH1FileHeadline = "H1-Errors"
    self._ratesL2FileHeadline = "L2-Errors, experimental order of convergence"
    self._ratesH1semiFileHeadline = "H1semi-Errors, experimental order of convergence"
    self._ratesH1FileHeadline = "H1-Errors, experimental order of convergence"
    
    
    self._maxErrL2 = None
    self._maxErrH1semi = None
    self._maxErrH1 = None

    self._summaryFileHeadline = "Simulation summary."
    self._summaryFileHeader = ["time step", "L2-Error", "Exp. L2-Order" \
      , "H1semi-Error", "Exp. H1semi-Order", "H1-Error", "Exp. H1-Order"]
    self._summaryFilename = "summary"

  def GetReady(self):
    from commics._tools.output import CreateNewFolder
    import sys
    
    if (self._firstRun):
      output_dir = CreateNewFolder(self.foldername)
      print("Folder --", output_dir, "-- for saving convergenceOrder results created.")
      self.foldername = output_dir
  
      self._integrators = []
      
      self._referenceSimulation.SetResultsFolder(self.foldername + "/simulation_reference")
  
      for j in range(self.R):
        self._integrators.append(self.SemiDeepCopy(self._referenceSimulation._integrator, self._otherParameters))
        self._integrators[j].SetParameter_TimeStepSize( \
          self._referenceSimulation.GetParameter_TimeStepSize() * 2**(j+1))
        self._integrators[j].SetResultsFolder(self.foldername \
          + "/simulation_" + str(2**(j+1)) + "xk")
  
      self._maxErrL2 = []
      self._maxErrH1semi = []
      self._maxErrH1 = []
      self._ratesL2 = []
      self._ratesH1semi = []
      self._ratesH1 = []
      for j in range(self.R):
        self._maxErrL2.append(0.0)
        self._maxErrH1semi.append(0.0)
        self._maxErrH1.append(0.0)
        self._ratesL2.append(0.0)
        self._ratesH1semi.append(0.0)
        self._ratesH1.append(0.0)
        
    else:
      for j in range(self.R):
        self._integrators[j].SetParameter_T_start(self._integrators[j].GetParameter_T_end())
        self._integrators[j].SetParameter_T_end(self._referenceSimulation.GetParameter_T_end())

    self._refTotalSteps = round( \
      (self._referenceSimulation._integrator.parameters.T_end - self._referenceSimulation._integrator.parameters.T_start) \
      / self._referenceSimulation._integrator.parameters.timeStepSize)

    if( self._refTotalSteps % (2**self.R) != 0 ):
      tmp = (2**self.R) - self._refTotalSteps % (2**self.R)
      self._refTotalSteps += tmp
      tmpTend = self._referenceSimulation.GetParameter_T_end()
      self._referenceSimulation.SetParameter_T_end( \
        self._referenceSimulation.GetParameter_T_start() \
        + self._refTotalSteps*self._referenceSimulation.GetParameter_TimeStepSize())

      warnString = "\n\n##### Warning: Changed T_end from "+ str(tmpTend) \
        + " to " + str(self._referenceSimulation.GetParameter_T_end()) \
        + " s.t. all simulations can finish at the same time!" + "\n" \
        + "Therefore this simulation should run without problems, " \
        + "but any reuse of the wrapper, i.e., calling GetReady() again " \
        + "may produce incorrect times in the output files!" \
        + " To avoid this problem, reconsider your combination of" \
        + " <T_start, T_end, timeStepSize, nrSimulations> #####" + "\n" +"\n"

      self.header += warnString
      self._summaryFileHeadline += warnString
      print(warnString)
      

      if( self._refTotalSteps % (2**self.R) != 0 ):
        print("Something went wrong. This case should never happen.")
        print("Check T_start, T_end, k you have chosen.")
        sys.exit()

    self._referenceSimulation._integrator.GetReady()
    for j in range(self.R):
      self._integrators[j]._integrator.GetReady()
      
    if (self._firstRun):
      self._InitializeErrorFiles()

  def Do(self):
    self.SaveParameters()
    if( self._firstRun ):
      self._referenceSimulation._integrator.Write()
      if( self._referenceSimulation._integrator.output.video ):
        self._referenceSimulation._integrator.SaveMagnetization()
      for j in range(self.R):
        self._integrators[j]._integrator.Write()
        if( self._integrators[j]._integrator.output.video ):
          self._integrators[j]._integrator.SaveMagnetization()

    for refStepCounter in range(1,self._refTotalSteps+1):
      self.StepAndWrite( -1 )
      
      for j in range(self.R):
        if( (refStepCounter % (2**(j+1)) ) == 0 ):
          self.StepAndWrite( j )

          if( j == self.R-1 ):
            self._saveErrL2Evolution()
            self._saveErrH1semiEvolution()
            self._saveErrH1Evolution()
            self._ComputeL2Rates()
            self._ComputeH1semiRates()
            self._ComputeH1Rates()
            self._SaveL2Rates()
            self._SaveH1semiRates()
            self._SaveH1Rates()

            self._PrintAllRates()

    self._firstRun = False
    
  
  def _ComputeL2Rates(self):
    from math import log
    for j in range(0,self.R-1):
      self._ratesL2[j] = log(self._maxErrL2[j+1]/self._maxErrL2[j]) / log(2)

  def _ComputeH1semiRates(self):
    from math import log
    for j in range(0,self.R-1):
      self._ratesH1semi[j] = log(self._maxErrH1semi[j+1]/self._maxErrH1semi[j]) / log(2)

  def _ComputeH1Rates(self):
    from math import log
    for j in range(0,self.R-1):
      self._ratesH1[j] = log(self._maxErrH1[j+1]/self._maxErrH1[j]) / log(2)

  def _SaveL2Rates(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._ratesL2filename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._ratesL2 \
      , nrwidth = self._columnWidth )
    
  def _SaveH1semiRates(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._ratesH1semifilename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._ratesH1semi \
      , nrwidth = self._columnWidth )

  def _SaveH1Rates(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._ratesH1filename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._ratesH1 \
      , nrwidth = self._columnWidth )

  def _PrintAllRates(self):
    lineSeperation = ""
    bottomLine = ""
    for j in range(self.R+1):
      lineSeperation += "-".ljust(self._shellColumnWidth,'-') + "|"
      bottomLine += "-".ljust(self._shellColumnWidth,'-') + "-"

    printString = "\n" + "\n" + "\n" + bottomLine + "\n"
    printString += ( "Simulation info at t = %.3e" \
      % self._referenceSimulation.GetTime() ).center( \
      (self.R+1)*(self._shellColumnWidth+1) - 1) + "|" 
    printString += "\n" + bottomLine + "\n"

    printString += " step size".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += (str(2**(j+1))+" x k ").center(self._shellColumnWidth) + "|"


    printString += "\n" + lineSeperation + "\n"
    printString += " L2-Error".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3e" % self._maxErrL2[j] ).center(self._shellColumnWidth) + "|"

    printString += "\n"
    printString += " Exp. Order ".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3f" % self._ratesL2[j] ).center(self._shellColumnWidth) + "|"


    printString += "\n" + lineSeperation + "\n"
    printString += " H1semi-Error".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3e" % self._maxErrH1semi[j] ).center(self._shellColumnWidth) + "|"

    printString += "\n"
    printString += " Exp. Order ".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3f" % self._ratesH1semi[j] ).center(self._shellColumnWidth) + "|"


    printString += "\n" + lineSeperation + "\n"
    printString += " H1-Error".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3e" % self._maxErrH1[j] ).center(self._shellColumnWidth) + "|"

    printString += "\n"
    printString += " Exp. Order ".ljust(self._shellColumnWidth) + "|"
    for j in range(self.R):
      printString += ( "%.3f" % self._ratesH1[j] ).center(self._shellColumnWidth) + "|"


    printString += "\n" + bottomLine + "\n" +"\n" +"\n"
    print(printString)

  def _saveErrL2Evolution(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._errL2filename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._maxErrL2 \
      , nrwidth = self._columnWidth )

  def _saveErrH1semiEvolution(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._errH1semifilename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._maxErrH1semi \
      , nrwidth = self._columnWidth )

  def _saveErrH1Evolution(self):
    from commics._tools.output import LiveSaveValues
    LiveSaveValues( self.foldername + "/" + self._errH1filename \
      + "_" + self._evolutionName + ".dat" \
      , self._referenceSimulation.GetTime(), self._maxErrH1 \
      , nrwidth = self._columnWidth )

  def StepAndWrite(self, who):
    import sys
    from commics._tools.output import LiveSaveValues
    if( type(who) != int or who < -1 or who > self.R-1 ):
      print("Error in StepAndWrite(who) --> who = ", who, "is not an int or not in range!")
      sys.exit()
    
    I = None
    if( who == -1):
      I = self._referenceSimulation._integrator
    else:
      I = self._integrators[who]._integrator
    
    I.Step()

    if( who != -1 ):
      errL2 = self._ComputeErrorL2(I)
      self._maxErrL2[who] = max( errL2, self._maxErrL2[who] )
      LiveSaveValues( I.output.foldername + "/" \
        + self._errL2filename + ".dat" \
        , I.Time(), [self._maxErrL2[who], errL2] \
        , nrwidth = self._columnWidth )

      errH1semi = self._ComputeErrorH1semi(I)
      self._maxErrH1semi[who] = max( errH1semi, self._maxErrH1semi[who] )
      LiveSaveValues( I.output.foldername + "/" \
        + self._errH1semifilename + ".dat" \
        , I.Time(), [self._maxErrH1semi[who], errH1semi] \
        , nrwidth = self._columnWidth )

      errH1 = self._ComputeErrorH1(I)
      self._maxErrH1[who] = max( errH1, self._maxErrH1[who] )
      LiveSaveValues( I.output.foldername + "/" \
        + self._errH1filename + ".dat" \
        , I.Time(), [self._maxErrH1[who], errH1] \
        , nrwidth = self._columnWidth )

  def _ComputeErrorL2(self, anIntegrator):
    from math import sqrt
    from ngsolve import Integrate
    err = 0.0
    
    m = anIntegrator._Mag.cf
    m_ref = self._referenceSimulation._integrator._Mag.cf

    err += Integrate( (m-m_ref)*(m-m_ref), anIntegrator._Mag._mesh )
    err *= (self._referenceSimulation.GetParameter_Scaling()**3) \
      * (self._referenceSimulation.GetParameter_Ms()**2)
    return sqrt(err)

  def _ComputeErrorH1semi(self, anIntegrator):
    from math import sqrt
    from ngsolve import Integrate
    from commics._tools.algebra import FrobeniusInnerProd
    err = 0.0
    
    Dm = anIntegrator._Mag.Dm
    Dm_ref = self._referenceSimulation._integrator._Mag.Dm

    err += Integrate( FrobeniusInnerProd(Dm-Dm_ref, Dm-Dm_ref), anIntegrator._Mag._mesh )
    
    err *= self._referenceSimulation.GetParameter_Scaling() \
      * (self._referenceSimulation.GetParameter_Ms()**2)

    return sqrt(err)

  def _ComputeErrorH1(self, anIntegrator):
    from math import sqrt
    err = self._ComputeErrorL2(anIntegrator)**2 \
      + self._ComputeErrorH1semi(anIntegrator)**2    

    return sqrt(err)


  def SemiDeepCopy(self, source, parameters):
    from commics import Integrator

    newIntegrator = Integrator( \
      self._otherScheme, \
      source._geometry, \
      parameters.DeepCopy(), \
      source.output.DeepCopy(), \
      source._numthreads)

    return newIntegrator

  def SaveParameters(self):
    from commics._tools.output import SaveContent
    saveString = ""
    if (self._firstRun):
      self._referenceSimulation.SaveMagnetization("mag_initial")
      self.SaveMesh()

      saveString += self.header
      saveString += "\n" + "\n"
      saveString += "Simulations for ( "
      for j in range(self.R):
        saveString += str(2**(j+1)) + " "
      saveString += ") x k_ref."
    
    saveString += "\n" + "\n"
    saveString += "Reference Simulation parameters:" + "\n" + "\n"
    saveString += str(self._referenceSimulation._integrator.parameters)
    saveString += "\n" + "\n" + "Geometry:" + "\n"
    saveString += str(self._referenceSimulation._integrator._geometry)
    saveString += "\n" + "\n" + "Time stepping:" + "\n"
    saveString += str(self._referenceSimulation)
    saveString += "\n" + "\n"
    saveString += "########################################"
    saveString += "########################################"
    saveString += "\n" + "\n"

    saveString += "Parameters for other simulations: " + "\n" + "\n"
    saveString += str(self._otherParameters)

    SaveContent(self.foldername + "/" + self.filenameParameters, saveString)
   
    self._referenceSimulation._integrator.SaveParameters()
    for j in range(self.R):
      self._integrators[j]._integrator.SaveParameters()

  def SaveMesh(self):
    self._referenceSimulation.SaveMesh(self.foldername + "/" + "mesh.vol")


  def _InitializeErrorFiles(self):
    writeString = "# " + self._errL2FileHeadline + "\n" + "# "
    for k in range(len(self._errL2FileHeader)):
      writeString = writeString + self._errL2FileHeader[k].ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    for j in range(self.R):
      newfile = open(self._integrators[j]._integrator.output.foldername + "/" \
        + self._errL2filename + ".dat", "a")
      newfile.write(writeString)
      newfile.close()


    writeString = "# " + self._errH1semiFileHeadline + "\n" + "# "
    for k in range(len(self._errH1semiFileHeader)):
      writeString = writeString + self._errH1semiFileHeader[k].ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    for j in range(self.R):
      newfile = open(self._integrators[j]._integrator.output.foldername + "/" \
        + self._errH1semifilename + ".dat", "a")
      newfile.write(writeString)
      newfile.close()


    writeString = "# " + self._errH1FileHeadline + "\n" + "# "
    for k in range(len(self._errH1FileHeader)):
      writeString = writeString + self._errH1FileHeader[k].ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    for j in range(self.R):
      newfile = open(self._integrators[j]._integrator.output.foldername + "/" \
        + self._errH1filename + ".dat", "a")
      newfile.write(writeString)
      newfile.close()
    

    writeString = "# " + self._errL2FileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._errL2filename \
      + "_" + self._evolutionName + ".dat", "a") 
    newfile.write(writeString)
    newfile.close()
    

    writeString = "# " + self._errH1semiFileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._errH1semifilename \
      + "_" + self._evolutionName + ".dat", "a") 
    newfile.write(writeString)
    newfile.close()


    writeString = "# " + self._errH1FileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._errH1filename \
      + "_" + self._evolutionName + ".dat", "a") 
    newfile.write(writeString)
    newfile.close()


    writeString = "# " + self._ratesL2FileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._ratesL2filename \
      + "_" + self._evolutionName + ".dat", "a")
    newfile.write(writeString)
    newfile.close()


    writeString = "# " + self._ratesH1semiFileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._ratesH1semifilename \
      + "_" + self._evolutionName + ".dat", "a")
    newfile.write(writeString)
    newfile.close()


    writeString = "# " + self._ratesH1FileHeadline + "\n"
    writeString += "# time".ljust(self._columnWidth)
    for j in range(self.R):
      writeString += (str(2**(j+1))+" x k").ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    newfile = open(self.foldername + "/" + self._ratesH1filename \
      + "_" + self._evolutionName + ".dat", "a")
    newfile.write(writeString)
    newfile.close()


  def Summarize(self):
    writeString = "# " + self._summaryFileHeadline + "\n" + "# "
    for k in range(len(self._summaryFileHeader)):
      writeString = writeString + self._summaryFileHeader[k].ljust(self._columnWidth) + " "

    writeString = writeString + "\n"
    
    for j in range(self.R):
      writeString += ( str(2**(j+1)) + " x k ").ljust(self._columnWidth) + " "
      writeString += str(self._maxErrL2[j]).ljust(self._columnWidth) + " "
      writeString += str(self._ratesL2[j]).ljust(self._columnWidth) + " "
      writeString += str(self._maxErrH1semi[j]).ljust(self._columnWidth) + " "
      writeString += str(self._ratesH1semi[j]).ljust(self._columnWidth) + " "
      writeString += str(self._maxErrH1[j]).ljust(self._columnWidth) + " "
      writeString += str(self._ratesH1[j]).ljust(self._columnWidth) + " "
      writeString += "\n"

    newfile = open(self.foldername + "/" \
      + self._summaryFilename + ".dat", "a")
    newfile.write(writeString)
    newfile.close()

  
  def SetMagFromReference(self):
    for j in range(self.R):
      self._integrators[j]._integrator._Mag.SetFromOther( self._referenceSimulation._integrator._Mag )

  
  def SetICFromIntegrator(self, other):
    self._referenceSimulation._integrator._Mag.SetFromOther( other._integrator._Mag )
    self.SetMagFromReference()

  def ProjectIC(self):
    self._referenceSimulation._integrator._Mag.Project()
    for j in range(self.R):
      self._integrators[j]._integrator._Mag.Project()


  def SetSolveTol(self, factor=1.0):
    self._referenceSimulation._integrator._solvetol = factor * self._referenceSimulation._integrator.parameters.TOL
    for j in range(self.R):
      self._integrators[j]._integrator._solvetol = factor * self._integrators[j]._integrator.parameters.TOL

  def SetResultsFolder(self, value):
    self.foldername = value

################################################################################
