from . import _interface
from abc import abstractmethod

################################################################################


class ComputeInterface(_interface._Interface):


#------------------------------------------------------------------------------#
  

  def GetReady(self):

    self._CheckForBadParameterCombinations()

    self._PhysicalToMathematical()

    if(self._counter == 0):
      self._geometry.GetReady()
      self._SetUpSpaces()
      self._InitializeMagnetization()
      self._InitializeOerstedField()
      self._SetUpAuxiliaryFunctions()
      self._CreateOutputFolder()
    elif(self._maxwellCouplingChanged):
      self._UpdateMaxwellCoupling()
    else:
      self._UpdateOerstedFieldFlags()
    self._maxwellCouplingChanged = False

    self._SetUpOutput()

    self._SetUpSystems()


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    self._Mag.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
      , self._solvetol)

    self._hc = None
    from commics._gridFunctions import oerstedField, _zeroField
    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc = _zeroField.ZeroField()


#------------------------------------------------------------------------------#


  def _UpdateOerstedFieldFlags(self):
    try:
      self._hc.UpdateFlags(self._flag_je_stationary)
    except(AttributeError):
      pass


#------------------------------------------------------------------------------#


  def _SetUpSpaces(self):
    from ngsolve import H1, VectorH1, ORDER_POLICY
    self._V = H1(self._geometry.GetMesh(), order=1)
    self._X = VectorH1(self._geometry.GetMesh(), order=1 \
      , order_policy=ORDER_POLICY.CONSTANT)

    self._psi = self._X.TrialFunction()
    self._phi = self._X.TestFunction()

    self._Dpsi = self._psi.Deriv()
    self._Dphi = self._phi.Deriv()



#------------------------------------------------------------------------------#


  def _InitializeMagnetization(self):
    from commics._gridFunctions import Magnetization

    self._Mag = Magnetization(self._geometry, self._V, self._X \
      , self.parameters.m0, self.parameters.maxwellCoupling, self._solvetol)


#------------------------------------------------------------------------------#


  def _InitializeOerstedField(self):
    from commics._gridFunctions import oerstedField, _zeroField

    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc = _zeroField.ZeroField()


#------------------------------------------------------------------------------#


  def Step(self):

    self._StepPrepare()
    self._StepSolve()
    self._StepUpdate()
    self._StepFollowUp()
  
  
#------------------------------------------------------------------------------#


  def _StepPrepare(self):
    from ngsolve import TaskManager

    self._StepPrepareBureaucracy()

#    with TaskManager(pajetrace=100*1000*1000):
    with TaskManager():
      self._StepPrepareWithTaskManager()

    self._StepPrepareWithoutTaskManager()



#------------------------------------------------------------------------------#


  def _StepPrepareBureaucracy(self):
    import time
    self._compTimer = time.time()
    
    self._ExistsMagnetization()

    self._ResetIterationCounter()
    self._ResetFixedPointIterationCounter()


#------------------------------------------------------------------------------#


  def _AssembleSystem(self):
    self._AssembleLHS()
    self._AssembleRHS()


#------------------------------------------------------------------------------#


  def _ComputeMaxwell(self):

    self._maxwellTimer.Start()
    self._Mag.ComputeMaxwell(self._solvetol)
    self._hc.Compute(self._solvetol)
    self._maxwellTimer.Stop()


#------------------------------------------------------------------------------#


  def _StepFollowUp(self):
    import time
    self._compTimeNow = time.time()-self._compTimer
    self._compTimeTotal += self._compTimeNow

    self._Mag.PrintAverages(self.Time())
    
    if (self.output.draw):
      self._Mag.Draw()
    
    self._counter += 1
    self._t += self._k
    self._time_cf.Set( self._t ) 

    self.Write()


#------------------------------------------------------------------------------#

  
  def Integrate(self, duration=None, relax=False):

    tmp_H_ext = self.parameters.H_ext
    tmp_maxwellCoupling = self.parameters.maxwellCoupling
    tmp_spintronicsCoupling = self.parameters.spintronicsCoupling
    tmp_T_end = self.parameters.T_end

    if relax:
      self.parameters.H_ext = None
      self.SetParameter_maxwellCoupling("strayField")
      self.parameters.spintronicsCoupling = "none"

    if duration != None:
      if self.parameters.T_start == None:
        self.parameters.T_start = 0.0
      elif self._counter != 0:
        self.parameters.T_start = self.Time()
      self.parameters.T_end = self.parameters.T_start + duration
          
    self.GetReady()
    self.Do()

    self.parameters.H_ext = tmp_H_ext
    self.SetParameter_maxwellCoupling(tmp_maxwellCoupling)
    self.parameters.spintronicsCoupling = tmp_spintronicsCoupling

    try:
      self.parameters.T_end = max(self.Time(), tmp_T_end)
    except:
      self.parameters.T_end = self.Time()


#------------------------------------------------------------------------------#


  def Do(self):
    self.SaveParameters()
    # TODO also a Debug of SaveParameters, for the math constants

    self._startCounter = self._counter
    if self._counter == 0: self.Write()

    if (self.output.draw): self._Mag.Draw()
    
    for step in range(self._steps):
      self.Step()
    
    print("Results have been written to:", self.output.foldername)
  
  
#------------------------------------------------------------------------------#


  @abstractmethod
  def _SetUpAuxiliaryFunctions(self): pass


#------------------------------------------------------------------------------#


  @abstractmethod
  def _SetUpSystems(self): pass


#------------------------------------------------------------------------------#


  @abstractmethod
  def _StepPrepareWithTaskManager(self): pass


#------------------------------------------------------------------------------#


  @abstractmethod
  def _StepPrepareWithoutTaskManager(self): pass


#------------------------------------------------------------------------------#


  @abstractmethod
  def _StepSolve(self): pass
    

#------------------------------------------------------------------------------#


  @abstractmethod
  def _StepUpdate(self): pass
    

#------------------------------------------------------------------------------#


################################################################################
