from . import _preconditionerInterface

################################################################################


class TPSPreconditioner(_preconditionerInterface._PreconditionerInterface):


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner(self):
    self._SetUpLHSPreconditioner()

    if(self.parameters.preconditionerType == "jacobi"):
      self._SetUpPreconditioner_Jacobi()
    elif(self.parameters.preconditionerType == "practical"):
      self._SetUpPreconditioner_Practical()
    elif(self.parameters.preconditionerType == "stationary"):
      self._SetUpPreconditioner_Stationary()
    elif(self.parameters.preconditionerType == None):
      self._SetUpPreconditioner_None()
    else:
      raise ValueError("preconditionerType = " \
      , self.parameters.preconditionerType, "not known")


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner(self):
    if(self.parameters.preconditionerType == "jacobi"):
      self._UpdatePreconditioner_Jacobi()
    elif(self.parameters.preconditionerType == "practical"):
      self._UpdatePreconditioner_Practical()
    elif(self.parameters.preconditionerType == "stationary"):
      self._UpdatePreconditioner_Stationary()
    elif(self.parameters.preconditionerType == None):
      self._UpdatePreconditioner_None()
    else:
      raise ValueError("preconditionerType = ", self.parameters.preconditionerType \
        , "not known")


#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    self._alphaPreconditioner = None
    self._a_stat_precond = None
    self._A_stat_precond = None
    self._preJacobi = None
    self._prePractical = None
    self._preStationary = None


#------------------------------------------------------------------------------#


  def _SetUpLHSPreconditioner(self):
    if(self.parameters.preconditionerType == None \
      or self.parameters.alphaPreconditioner == None):
      return

    from commics._tools import ScipySparseFromNgMatrix
    from ngsolve import TaskManager, SymbolicBFI, BilinearForm

    lhs_stat_precond_cf = self._MassTermPrecond() + self._StiffTerm()

    self._a_stat_precond = BilinearForm(self._X)
    self._a_stat_precond += SymbolicBFI( \
      lhs_stat_precond_cf.Compile(realcompile=True, wait=True))
    
    with TaskManager():
      self._a_stat_precond.Assemble()
    self._A_stat_precond = ScipySparseFromNgMatrix(self._a_stat_precond)

  
#------------------------------------------------------------------------------#


  def _MassTermPrecond(self):
    return self.parameters.alphaPreconditioner * self._Mass()


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Jacobi(self):
    import scipy.sparse.linalg
    if(self.parameters.alphaPreconditioner == None):
      qjacq = 1.0 / self._A_stat.diagonal()[:self._A_stat.shape[0]//3]
    else:
      qjacq = 1.0 \
        / self._A_stat_precond.diagonal()[:self._A_stat_precond.shape[0]//3]

    self._preJacobi = qjacq.repeat(2)
    
    jacobi2D = lambda x: self._preJacobi*x
    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      (self._preJacobi.size, self._preJacobi.size), matvec=jacobi2D)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Practical(self):
    import scipy.sparse.linalg
    import scipy.sparse as sp
    import numpy as np
    import pyamg

    if(self.parameters.alphaPreconditioner == None):
      A = self._A_stat[:self._A_stat.shape[0]//3, :self._A_stat.shape[0]//3]
    else:
      A = self._A_stat_precond[:self._A_stat_precond.shape[0]//3 \
        , :self._A_stat_precond.shape[0]//3]

    self._Compute_Q()

    # realize "inverse" of stationary part via multilevel solver
    ml = pyamg.ruge_stuben_solver(A)
    pre = ml.aspreconditioner()
    pre1D = lambda z: pre * z
  
    self._prePractical = lambda y: \
      pre1D(y.reshape((y.size//3,3), order='F')).reshape(y.size, order='F')

    practical2D = lambda x: \
      self._Q.dot(self._prePractical(self._Q.transpose().dot(x)))

    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      (self._Q.shape[0], self._Q.shape[0]), matvec=practical2D)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Stationary(self):
    import scipy.sparse.linalg
    import scipy.sparse as sp
    import numpy as np
    import pyamg

    if(self.parameters.alphaPreconditioner == None):
      A = self._A_stat[:self._A_stat.shape[0]//3, :self._A_stat.shape[0]//3]
    else:
      A = self._A_stat_precond[:self._A_stat_precond.shape[0]//3 \
        , :self._A_stat_precond.shape[0]//3]

    # realize "inverse" of stationary part via multilevel solver
    ml = pyamg.ruge_stuben_solver(A)
    pre = ml.aspreconditioner()
    pre1D = lambda z: pre * z

    stationary2D = lambda x: \
      pre1D(x.reshape((x.size//2,2), order='C')).reshape(x.size, order='C')
    
    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      (A.shape[0]*2, A.shape[0]*2), matvec=stationary2D)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_None(self):
    self._preconditioner = None


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Jacobi(self):
    qjacq = 1.0 / self._LHS_3D.diagonal()[:self._LHS_3D.shape[0]//3]
    self._preJacobi = qjacq.repeat(2)


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Practical(self):
    pass


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Stationary(self):
    pass


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_None(self):
    pass


#------------------------------------------------------------------------------#


################################################################################

