### KW1 ###
from . import _integrator
from . import _forms


################################################################################


class KW1(_integrator._Integrator, \
  _forms.lf_derivative.LinearForms_Derivative):


#------------------------------------------------------------------------------#

  
  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):

    _integrator._Integrator.__init__(self \
      , geometry, parameters, output, numthreads)

    self._theta = 0.5

    self._maxiter = 20


#------------------------------------------------------------------------------#


  def GetReady(self):

    self._itertol = self.parameters.TOL
    self._solvetol = self.parameters.TOL / 100.0
    self._theta = self.parameters.theta

    _integrator._Integrator.GetReady(self)


#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    from commics._gridFunctions import Derivative, Magnetization

    self._Mag.v = Derivative(self._GetMesh(), self._V, self._X \
      , self.parameters.maxwellCoupling, self._solvetol)

    self._MagUnprojected = Magnetization(self._geometry, self._V, self._X \
      , self._Mag, maxwellCoupling="none", solvetol=self._solvetol)


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    _integrator._Integrator._UpdateMaxwellCoupling(self)

    self._Mag.v.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
      , self._solvetol)

    self._hc_impl = None
    from commics._gridFunctions import oerstedField, _zeroField
    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc_impl = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_impl_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc_impl = _zeroField.ZeroField()


#------------------------------------------------------------------------------#

  
  def _InitializeOerstedField(self):
    super()._InitializeOerstedField()

    from commics._gridFunctions import oerstedField, _zeroField
    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc_impl = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_impl_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc_impl = _zeroField.ZeroField()


#------------------------------------------------------------------------------#


  def _UpdateOerstedFieldFlags(self):
    super()._UpdateOerstedFieldFlags()
    try:
      self._hc_impl.UpdateFlags(self._flag_je_stationary)
    except(AttributeError):
      pass


#------------------------------------------------------------------------------#


  def _ComputeMaxwell(self):
    super()._ComputeMaxwell()

    self._maxwellTimer.Start()
    self._hc_impl.Compute(self._solvetol)
    self._maxwellTimer.Stop()


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    from commics._tools import ScipySparseFromNgMatrix
    from ngsolve import IntegrationRule, BilinearForm, SymbolicBFI \
      , TaskManager, LinearForm, SymbolicLFI
    import numpy as np
    import scipy.sparse as sp

    massLumping = IntegrationRule( \
      points = [(0,0,0), (1,0,0), (0,1,0), (0,0,1)] \
      , weights = [1/24, 1/24, 1/24, 1/24] )

    ### Sh begin
    sh_cf = self._KW_SkewTerm()

    self._sh = BilinearForm(self._X)
    self._sh += SymbolicBFI(sh_cf.Compile(realcompile=True, wait=True) \
      , intrule=massLumping)

    with TaskManager():
      self._sh.Assemble()
    self._Sh = ScipySparseFromNgMatrix(self._sh)
    ### Sh end

    ### Mh Mh^-1 begin
    mh_cf = self._KW_MassTerm()

    self._mh = BilinearForm(self._X)
    self._mh += SymbolicBFI(mh_cf.Compile(realcompile=True, wait=True) \
      , intrule=massLumping)

    with TaskManager():
      self._mh.Assemble()
    self._Mh = ScipySparseFromNgMatrix(self._mh)

    self._Mh_diag = self._Mh.diagonal()
    self._MhInv_diag = 1.0 / self._Mh_diag
    self._MhInv = sp.csr_matrix((self._MhInv_diag \
      , (np.arange(self._MhInv_diag.size), np.arange(self._MhInv_diag.size))))

    self._Mh.eliminate_zeros()
    ### Mh Mh^-1 end

    ### M begin
    m_cf = self._KW_MassTerm()

    self._m = BilinearForm(self._X)
    self._m += SymbolicBFI(m_cf.Compile(realcompile=True, wait=True))

    with TaskManager():
      self._m.Assemble()
    self._M = ScipySparseFromNgMatrix(self._m)
    self._M.eliminate_zeros()
    ### Mh end

    ### A begin
    a_cf = self._KW_StiffTerm()

    self._a = BilinearForm(self._X)
    self._a += SymbolicBFI(a_cf.Compile(realcompile=True, wait=True))

    with TaskManager():
      self._a.Assemble()
    self._A = ScipySparseFromNgMatrix(self._a)
    self._A.eliminate_zeros()
    ### A end

    ### DiscreteLaplacian begin
    self._DiscreteLaplacianMatrix = - self._MhInv * self._A
    ### DiscreteLaplacian end

#    ### DiscreteProjection begin
#    self._DiscreteProjectionMatrix = self._MhInv * self._M
#    ### DiscreteProjection end

    self._LHS_stat = self._Mh

    ### lf_fix begin
    rhs_fix_cf = self._KW_ExchangeTerm()
    rhs_fix_cf -= self._LinearTerms_pi()
    rhs_fix_cf -= self._NonLinearTerms_Pi()
    rhs_fix_cf -= self._KW_ExplicitTerms()

    self._f_fix_parts = LinearForm(self._X)
    self._f_fix_parts += SymbolicLFI(rhs_fix_cf.Compile( \
      realcompile=True, wait=True))
    ### lf_fix end

    ### lf_iter begin
    rhs_iter_cf = -self._DerivLinearTerms_pi()
    rhs_iter_cf -= self._DerivNonLinearTerms_Pi()

    self._f_iter_parts = LinearForm(self._X)
    self._f_iter_parts += SymbolicLFI(rhs_iter_cf.Compile( \
      realcompile=True, wait=True))
    ### lf_iter end


#------------------------------------------------------------------------------#

  
  def _Sh_MhInv(self, x):
    return self._Sh * (self._MhInv_diag * x)


#------------------------------------------------------------------------------#


  def _StepPrepareWithTaskManager(self):

    self._ComputeMaxwell()
    self._Assemble_Sh()
    self._AssembleFixPartsRHS()


#------------------------------------------------------------------------------#


  def _StepPrepareWithoutTaskManager(self): pass


#------------------------------------------------------------------------------#


  def _Assemble_Sh(self):
    self._sh.Assemble()
    self._Sh.data = self._sh.mat.COO()[2].NumPy()


#------------------------------------------------------------------------------#


  def _AssembleFixPartsRHS(self):
    self._f_fix_parts.Assemble()


#------------------------------------------------------------------------------#


  def _AssembleIterPartsRHS(self):
    self._f_iter_parts.Assemble()


#------------------------------------------------------------------------------#


  def _StepSolve(self):
    self._Compute_v()
    

#------------------------------------------------------------------------------#


  def _Compute_LHS_instat(self):
    self._LHS_instat = -self.parameters.alpha * self._Sh      

    if self._theta != 0.0:
      self._LHS_instat += (self._C_ex * self._k * self._theta) \
        * (self._Sh * self._DiscreteLaplacianMatrix)


#------------------------------------------------------------------------------#


  def _Compute_RHS_fix(self):
    self._RHS_fix = self._Sh_MhInv(self._f_fix_parts.vec.FV().NumPy())


#------------------------------------------------------------------------------#


  def _Compute_RHS_iter(self):
    from ngsolve import TaskManager 

    with TaskManager():
      self._AssembleIterPartsRHS()

    self._RHS_iter = self._Sh_MhInv(self._f_iter_parts.vec.FV().NumPy())


#------------------------------------------------------------------------------#


  def _Compute_v(self):
    import sys
    from scipy import sqrt
    import scipy.sparse.linalg
    from ngsolve import Integrate, GridFunction
    import numpy as np

    if( self.parameters.maxwellCoupling == "magnetostaticField" \
      or self.parameters.maxwellCoupling == "strayField" ):
      self._Mag.v.hs.SetZero()
    
    # assemble blf and lf
    self._Compute_LHS_instat()
    
    LHS = self._LHS_stat + self._LHS_instat
    
    itnr = 0 
    oldIterate = np.array( self._Mag.v.gf.vec.FV() )
    
    diff_gf = GridFunction(self._X)
    diff_cf = diff_gf
    
    self._Compute_RHS_fix()

    for itnr in range(self._maxiter):
      self._Compute_RHS_iter()
      RHS = self._RHS_fix + self._RHS_iter

      self._Mag.v.gf.vec.FV().NumPy()[:], succ \
        = scipy.sparse.linalg.gmres(LHS, RHS , x0=None \
        , tol=self._solvetol, maxiter=4000, M=None)
      
    
      diff_gf.vec.FV().NumPy()[:] = self._Mag.v.gf.vec.FV().NumPy()-oldIterate
      itError = sqrt( Integrate(diff_cf*diff_cf, self._GetMesh()) )
    
      if (itError <= self._itertol):
        break
    
      print("itnr = ", itnr, "; Err = ", itError)
    
      oldIterate[:] = self._Mag.v.gf.vec.FV().NumPy()
      self._Mag.v.ComputeMaxwell(self._solvetol)

    print("itnr = ", itnr, "; Err = ", itError)
    

#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    self._Mag.m.vec.data += self._k * self._Mag.v.gf.vec

    if self.output.traceUnprojectedExchangeEnergy:
      self._MagUnprojected.SetFromOther(self._Mag, withStrayField=False)

    self._Mag.Project()
    

#------------------------------------------------------------------------------#

  
  def _KW_MassTerm(self):
    return self._Mass()


#------------------------------------------------------------------------------#


  def _KW_StiffTerm(self):
      return self._Stiff()


#------------------------------------------------------------------------------#


  def _KW_SkewTerm(self):
    return self._Skew(self._Mag.m)


#------------------------------------------------------------------------------#


  def _KW_ExchangeTerm(self):
    return self._C_ex * self._Exchange(self._Mag.Dm)


#------------------------------------------------------------------------------#


  def _KW_ExplicitTerms(self):
    return self._KW_AppliedTerm() + self._KW_OerstedFieldTerm_impl()


#------------------------------------------------------------------------------#


  def _KW_AppliedTerm(self):
    return self._Applied_impl()


#------------------------------------------------------------------------------#


  def _KW_OerstedFieldTerm_impl(self):
    return self._Oersted(self._hc_impl)


#------------------------------------------------------------------------------#


  def SetTheta(self, value):
    self.parameters.theta = value


#------------------------------------------------------------------------------#


  def TraceUnprojectedExchangeEnergy(self, value):
    self.output.traceUnprojectedExchangeEnergy = value


#------------------------------------------------------------------------------#


################################################################################
