### TPS2 ###
from . import _tpsx
from . import _forms


################################################################################


class TPS2(_tpsx.TPSX, \
  _forms.lf_derivative.LinearForms_Derivative, \
  _forms.blf_piImplicit.BilinearForms_piImplicit):
  '''
  implements integrator from
  Alouges, F., Kritsikis, E., Steiner, J., and Toussaint, J.-C. A convergent and precise finite element scheme for Landau–Lifschitz–Gilbert equation. Numer. Math. 128, 3 (2014), 407–430.
  '''


#------------------------------------------------------------------------------#

  
  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):

    _tpsx.TPSX.__init__(self, geometry, parameters, output, numthreads)

    from scipy import log
    self.rho = lambda q : q*abs(log(q))
    self.MWeightCut = lambda q : 1.0 / (q*abs(log(q)))

    self._theta = 0.5

    self._maxiter = 20


#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    from commics._gridFunctions import Magnetization

    super()._SetUpAuxiliaryFunctions()

    self._MagUnprojected = Magnetization(self._geometry, self._V, self._X \
      , self._Mag, maxwellCoupling="none", solvetol=self._solvetol)


#------------------------------------------------------------------------------#


  def GetReady(self):
    self._itertol = self.parameters.TOL
    self._solvetol = self.parameters.TOL / 100.0

    _tpsx.TPSX.GetReady(self)
    self._Mag.Project()


#------------------------------------------------------------------------------#


  def _CheckForBadParameterCombinations(self):
    if (self.parameters.theta != 0.5):
      raise ValueError("theta!=0.5 not available for", type(self))

    _tpsx.TPSX._CheckForBadParameterCombinations(self)


#------------------------------------------------------------------------------#

  
  def _InitializeOerstedField(self):
    super()._InitializeOerstedField()

    from commics._gridFunctions import oerstedField, _zeroField
    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc_impl = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_impl_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc_impl = _zeroField.ZeroField()


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):
    super()._UpdateMaxwellCoupling()

    self._hc_impl = None
    from commics._gridFunctions import oerstedField, _zeroField
    if self.parameters.spintronicsCoupling != "none" \
      and (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      self._hc_impl = oerstedField.OerstedField(self._geometry.GetMesh() \
        , self._j_impl_cf_oersted, self._solvetol, self._flag_je_stationary)
    else:
      self._hc_impl = _zeroField.ZeroField()


#------------------------------------------------------------------------------#


  def _UpdateOerstedFieldFlags(self):
    super()._UpdateOerstedFieldFlags()
    try:
      self._hc_impl.UpdateFlags(self._flag_je_stationary)
    except(AttributeError):
      pass


#------------------------------------------------------------------------------#


  def _ComputeMaxwell(self):
    super()._ComputeMaxwell()

    self._maxwellTimer.Start()
    self._hc_impl.Compute(self._solvetol)
    self._maxwellTimer.Stop()


#------------------------------------------------------------------------------#


  def _SetRHS_3D(self):
    self._RHS_3D = self._f_fix.vec.FV().NumPy() + self._f_iter.vec.FV().NumPy()


#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    self._Mag.m.vec.data += self._k * self._Mag.v.gf.vec

    if self.output.traceUnprojectedExchangeEnergy:
      self._MagUnprojected.SetFromOther(self._Mag, withStrayField=False)

    self._Mag.Project()


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    from commics._tools import ScipySparseFromNgMatrix
    from ngsolve import TaskManager, TET, IntegrationRule, SymbolicLFI, LinearForm, SymbolicBFI, BilinearForm

    #blf
    lhs_stat_cf = self._StiffTerm() + self._DerivLinearTerms_pi_Implicit_LHS()
    lhs_instat_cf = self._MassTerm() + self._SkewTerm()
    
    self._a_stat = BilinearForm(self._X)
    self._a_stat += SymbolicBFI( lhs_stat_cf.Compile(realcompile=True, wait=True) )
    
    with TaskManager():
      self._a_stat.Assemble()
    self._A_stat = ScipySparseFromNgMatrix(self._a_stat)
    
    ir3 = IntegrationRule(TET, order=3)
    
    self._a_instat = BilinearForm(self._X)
    self._a_instat += SymbolicBFI( lhs_instat_cf.Compile(realcompile=True, wait=True), intrule=ir3 )
    
    with TaskManager():
      self._a_instat.Assemble()
    self._A_instat = ScipySparseFromNgMatrix(self._a_instat) 

    #lf
    rhs_fix_cf = self._ExchangeTerm()
    rhs_fix_cf += self._LinearTerms_pi()
    rhs_fix_cf += self._NonLinearTerms_Pi()
    rhs_fix_cf += self._ExplicitTerms()

    self._f_fix = LinearForm(self._X)
    self._f_fix += SymbolicLFI(rhs_fix_cf.Compile(realcompile=True, wait=True))

    rhs_iter_cf = self._DerivLinearTerms_pi_Implicit_RHS()
    rhs_iter_cf += self._DerivNonLinearTerms_Pi()
    
    self._f_iter = LinearForm(self._X)
    self._f_iter += SymbolicLFI(rhs_iter_cf.Compile(realcompile=True, wait=True))


#------------------------------------------------------------------------------#


  def _ComputeWeight(self):

    tmp_cf = self._WeightExchange()

    if (self.parameters.H_ext != None and any(self.parameters.H_ext)):
      tmp_cf += self._WeightApplied()

    if (self._C_ani):
      tmp_cf += self._WeightUniaxialAnisotropy()

    if (self.parameters.dmCoupling != "none"):
      tmp_cf += self._WeightDzyaloshinskiiMoriya()

    if(self.parameters.maxwellCoupling != "none"):
      tmp_cf += self._WeightMaxwell()

    if(self.parameters.spintronicsCoupling != "none"):
      tmp_cf += self._WeightSpintronics()

    return self._WeightCut(tmp_cf)


#------------------------------------------------------------------------------#


  def _Compute_v(self):
    import sys
    from scipy import sqrt
    import scipy.sparse.linalg
    from ngsolve import Integrate, GridFunction
    import numpy as np
    if( self.parameters.maxwellCoupling == "magnetostaticField" \
      or self.parameters.maxwellCoupling == "strayField" ):
      self._Mag.v.hs.SetZero()
    
    itnr = 0 
    oldIterate = np.array( self._Mag.v.gf.vec.FV() )
    
    diff_gf = GridFunction(self._X)
    diff_cf = diff_gf
    
    # assemble blf and lf
    self._Assemble_f_fix()

    for itnr in range(self._maxiter):
      self._Assemble_f_iter()
      f = self._f_fix.vec.FV().NumPy() + self._f_iter.vec.FV().NumPy()

      RHS = self._Q.dot( f )
    
      self._UpdateGuess()
      self._solution, succ = scipy.sparse.linalg.gmres(self._LHS, RHS , x0=self._guess \
        , tol=self._solvetol, maxiter=4000, M=self._preconditioner)
      self._Mag.v.gf.vec.FV().NumPy()[:] = self._Q.transpose().dot(self._solution)
    
      diff_gf.vec.FV().NumPy()[:] = self._Mag.v.gf.vec.FV().NumPy()-oldIterate
      itError = sqrt( Integrate(diff_cf*diff_cf, self._GetMesh()) )
    
      if (itError <= self._itertol):
        break
    
      print("itnr = ", itnr, "; Err = ", itError)
    
      oldIterate[:] = self._Mag.v.gf.vec.FV().NumPy()
      self._Mag.v.ComputeMaxwell(self._solvetol)

    print("itnr = ", itnr, "; Err = ", itError)


#------------------------------------------------------------------------------#


  def _ComputeRHS(self): pass


#------------------------------------------------------------------------------#


  def _Assemble_f_fix(self):
    from ngsolve import TaskManager
    with TaskManager():
      self._f_fix.Assemble()


#------------------------------------------------------------------------------#


  def _Assemble_f_iter(self):
    from ngsolve import TaskManager
    with TaskManager():
      self._f_iter.Assemble()


#------------------------------------------------------------------------------#


  def _MassTerm(self):
    return self._ComputeWeight() * self._Mass()


#------------------------------------------------------------------------------#


  def _StiffTerm(self):
    return self._C_ex * self._theta * self._k * (1.0 + self.rho(self._k)) \
      * self._Stiff()


#------------------------------------------------------------------------------#


  def _AppliedTerm(self):
    return self._Applied_impl()


#------------------------------------------------------------------------------#


  def _OerstedFieldTerm(self):
    return self._Oersted(self._hc_impl)


#------------------------------------------------------------------------------#


  @staticmethod
  def _IsMagProjected():
    return True


#------------------------------------------------------------------------------#


  def _WeightExchange(self):
    from commics._tools.algebra import FrobeniusInnerProd
    return -self._C_ex * FrobeniusInnerProd(self._Mag.Dm, self._Mag.Dm)
    

#------------------------------------------------------------------------------#


  def _WeightApplied(self):
    from commics._tools.algebra import InnerProd
    return InnerProd(self._hext_cf, self._Mag.m)
    

#------------------------------------------------------------------------------#


  def _WeightUniaxialAnisotropy(self):
    from commics._tools.algebra import InnerProd
    return -self._C_ani \
      * InnerProd(self._Mag.phiUniaxialAnisotropy(self._easy_cf), self._Mag.cf)
    

#------------------------------------------------------------------------------#


  def _WeightDzyaloshinskiiMoriya(self):
    import sys
    from commics._tools.algebra import Curl
    from commics._tools.algebra import InnerProd
    if(self.parameters.dmCoupling == "bulk"):
      return -self._C_dm \
        * InnerProd(Curl(self._Mag.Dm), self._Mag.m)
    elif(self.parameters.dmCoupling == "interfacial"):
      from commics.integrators._forms._linearForms._opInterfacialDMI \
        import _OpInterfacialDMI
      return -self._C_dm \
        * InnerProd(self._Mag.m \
        , _OpInterfacialDMI(self._Mag.Dm, self.parameters.dmi_axis))
    elif (self.parameters.dmCoupling != "none"):
      print("Error in TPS2._ComputeWeight():")
      print("Dzyaloshinskii-Moriya options not >bulk<, >interfacial< \
        or >none< are not implemented yet")
      sys.exit()
    

#------------------------------------------------------------------------------#


  def _WeightMaxwell(self):
    import sys
    from commics._tools.algebra import InnerProd

    if (self.parameters.maxwellCoupling=="magnetostaticField"):
      return InnerProd(self._Mag.hs.cf, self._Mag.m) \
        + InnerProd(self._hc.cf, self._Mag.m)
    elif (self.parameters.maxwellCoupling=="strayField"):
      return InnerProd(self._Mag.hs.cf, self._Mag.m)
    elif (self.parameters.maxwellCoupling=="oerstedField"):
      return InnerProd(self._hc.cf, self._Mag.m)
    

#------------------------------------------------------------------------------#


  def _WeightSpintronics(self):
    import sys
    if(self.parameters.spintronicsCoupling == "zhangLi"):
      return self._WeightZhangLi()
    elif(self.parameters.spintronicsCoupling == "slonczewski"):
      return self._WeightSlonczewski()
    elif(self.parameters.spintronicsCoupling != "none"):
      print("Error in TPS2._ComputeWeight():")
      print("Spintronics options not >zhangLi<, >slonczewski< \
        or >none< are not implemented yet")
      sys.exit()


#------------------------------------------------------------------------------#


  def _WeightZhangLi(self):
    from commics._tools.algebra import CrossProd
    from ngsolve import CoefficientFunction
    return self._C_st * CoefficientFunction( CrossProd( self._Mag.m \
      , self._Mag.opZhangLi(self._j_cf) ) ) * self._Mag.cf \
      + self._C_st * self.parameters.xi \
      * CoefficientFunction(self._Mag.opZhangLi(self._j_cf) ) \
      * self._Mag.cf
    

#------------------------------------------------------------------------------#


  def _WeightSlonczewski(self):
    from commics._tools.algebra import CrossProd
    from ngsolve import CoefficientFunction
    from commics.integrators._forms._linearForms._gSlonczewski import _G
    G = _G(self._j_cf, self.parameters.P, self._p_cf, self._Mag.m)

    return self._C_st * G * CoefficientFunction( CrossProd( self._Mag.m \
      , self._p_cf ) ) * self._Mag.cf


#------------------------------------------------------------------------------#


  def _WeightCut(self, cf):
    from ngsolve import IfPos
    Min = lambda a, b : IfPos(a-b, b, a)

    case_pos = self.parameters.alpha + self._k*self._theta \
      * Min(cf, self.MWeightCut(self._k))
    case_neg = self.parameters.alpha \
      / (1.0 + self._k*self._theta/self.parameters.alpha \
      * Min(-cf, self.MWeightCut(self._k)))

    return IfPos(cf, case_pos, case_neg)


#------------------------------------------------------------------------------#


  def _AssembleRHS(self): pass


#------------------------------------------------------------------------------#


  def TraceUnprojectedExchangeEnergy(self, value):
    self.output.traceUnprojectedExchangeEnergy = value


#------------------------------------------------------------------------------#


################################################################################
