###  TPS2+AB ###
from . import tps2
from . import _forms


################################################################################


class TPS2AB(tps2.TPS2, _forms.lf_old.LinearForms_Old):
  '''
  implements IMEX approach from
  Di Fratta, G., Pfeiler, C.-M., Praetorius, D., Ruggeri, M., and Stiftner, B. Linear second order IMEX-type integrator for the (eddy current) Landau–Lifshitz–Gilbert equation. Submitted for publication, preprint available at arXiv:1711.10715, 2017.
  '''


#------------------------------------------------------------------------------#

  
  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):
    
    tps2.TPS2.__init__(self, geometry, parameters, output, numthreads)

    self._nowImplicit = True


#------------------------------------------------------------------------------#


  def GetReady(self):
    self._tsm_new = 1.0 + self._theta
    self._tsm_old = -self._theta
    
    tps2.TPS2.GetReady(self)
    self._nowImplicit = True


#------------------------------------------------------------------------------#


  def _SetRHS_3D(self):
    if self._nowImplicit:
      tps2.TPS2._SetRHS_3D(self)
    else:
      self._RHS_3D = self._ab_f.vec.FV().NumPy()


#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    tps2.TPS2._SetUpAuxiliaryFunctions(self)

    self._SetUpOldMagnetization()


#------------------------------------------------------------------------------#

  
  def _SetUpOldMagnetization(self):
    from commics._gridFunctions import Magnetization

    self._Mag_old = Magnetization(self._geometry, self._V, self._X \
      , self._Mag, self.parameters.maxwellCoupling, self._solvetol)


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    tps2.TPS2._UpdateMaxwellCoupling(self)

    self._Mag_old.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
      , self._solvetol)


#------------------------------------------------------------------------------#


  def _StepFollowUp(self):
    
    tps2.TPS2._StepFollowUp(self)

    self._nowImplicit = False


#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    self._Mag_old.SetFromOther(self._Mag)

    tps2.TPS2._StepUpdate(self)


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    from ngsolve import SymbolicLFI, LinearForm

    tps2.TPS2._SetUpSystems(self)

    #lf
    rhs_ab_cf = self._ExchangeTerm()
    rhs_ab_cf += self._ab_LinearTerms_pi()
    rhs_ab_cf += self._ab_NonLinearTerms_Pi()
    rhs_ab_cf += self._ExplicitTerms()

    self._ab_f = LinearForm(self._X)
    self._ab_f += SymbolicLFI(rhs_ab_cf.Compile(realcompile=True, wait=True))


#------------------------------------------------------------------------------#


  def _Compute_v(self):
    import scipy.sparse.linalg
    if(self._nowImplicit):
      tps2.TPS2._Compute_v(self)
      return

    self._solution, succ = scipy.sparse.linalg.gmres(self._LHS, self._RHS \
      , x0=self._guess, tol=self._solvetol, maxiter=4000 \
      , M=self._preconditioner, callback=self._callback)

    self._Mag.v.gf.vec.FV().NumPy()[:] = self._Q.transpose().dot(self._solution)

    self._UpdateGuess()
    

#------------------------------------------------------------------------------#


  def _ComputeRHS(self):
    from . import _tpsx
    _tpsx.TPSX._ComputeRHS(self)


#------------------------------------------------------------------------------#


  def _AssembleRHS(self):
    self._Assemble_f_ab()


#------------------------------------------------------------------------------#


  def _Assemble_f_ab(self):
    self._ab_f.Assemble()


#------------------------------------------------------------------------------#


  def _ab_LinearTerms_pi(self):
    return self._LinearTerms_pi_Implicit_RHS() \
      + self._tsm_new * self._LinearTerms_pi_Explicit_RHS() \
      + self._tsm_old * self._OldLinearTerms_pi_Explicit_RHS()


#------------------------------------------------------------------------------#


  def _ab_NonLinearTerms_Pi(self):
    return self._NonLinearTerms_Pi() \
      + (self._tsm_new - 1.0) * self._NowDerivNonLinearTerms() \
      + self._tsm_old * self._OldDerivNonLinearTerms()


#------------------------------------------------------------------------------#


################################################################################
